#include "OgreOctreeHierarchyInterface.h"
#include "OgreVisibilityOctreeSceneManager.h"
#include <OgreOctree.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>


namespace Ogre {

//-----------------------------------------------------------------------
OctreeHierarchyInterface::OctreeHierarchyInterface(OctreeSceneManager *sm, RenderSystem *rsys): 
SceneNodeHierarchyInterface(sm, rsys)
{
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::TraverseNode(GtpVisibility::HierarchyNode *node)
{
	++ mNumTraversedNodes;

	Octree *octree = static_cast<Octree *>(node);

	// if we come across some renderable geometry => render it
	if (octree->mNodes.size() > 0)
	{
		RenderNode(node);
	}
	
	//if (octree->numNodes() > (int)octree->mNodes.size()) // if not all subtrees are empty
	if (!IsLeaf(node))
	{
		for(int i=0; i<8; ++i)
		{
			Octree *nextChild = 
				octree->mChildren[(i & 4) >> 2][(i & 2) >> 1][i & 1];
			
			if (nextChild)
			{
				mDistanceQueue->push(nextChild);
			}
		}
	}
}
//-----------------------------------------------------------------------
bool OctreeHierarchyInterface::IsLeaf(GtpVisibility::HierarchyNode *node) const
{
	Octree *octree = static_cast<Octree *>(node);
	// HACK: if there are subtrees, they are empty => we are not interested in them
	return octree->numNodes() == (int)octree->mNodes.size();
}
//-----------------------------------------------------------------------
bool OctreeHierarchyInterface::HasGeometry(GtpVisibility::HierarchyNode *node) const
{
	return !(static_cast<Octree *>(node))->mNodes.empty();
}
//-----------------------------------------------------------------------
float OctreeHierarchyInterface::GetSquaredDistance(GtpVisibility::HierarchyNode *node) const
{
	LogManager::getSingleton().logMessage("here4");

	const AxisAlignedBox &box = static_cast<Octree *>(node)->mBox;
	const Vector3 pos = ((box.getMaximum() - box.getMinimum()) * 0.5) + box.getMinimum();
	
	return (mCullCamera->getDerivedPosition() - pos).squaredLength();
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::SetNodeVisible(GtpVisibility::HierarchyNode *node, 
											  const bool visible) const
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	static_cast<Octree *>(node)->setOctreeVisible(visible);
#endif
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::SetLastVisited(GtpVisibility::HierarchyNode *node, 
											  const unsigned int frameId) const
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	static_cast<Octree *>(node)->setLastVisited(frameId);
#endif
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::PullUpVisibility(GtpVisibility::HierarchyNode *node) const
{		
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	Octree *octant = static_cast<Octree *>(node);

	while (octant && !octant->isOctreeVisible())
	{
		octant->setOctreeVisible(true);
		octant = octant->getParent();
	}
#endif
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::RenderNode(GtpVisibility::HierarchyNode *node)
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	Octree *octant = static_cast<Octree *>(node);

	if (octant->lastRendered() != mFrameId)
	{
		octant->setLastRendered(mFrameId);

		dynamic_cast<OctreeSceneManager *>(mSceneManager)->_renderOctant(mCamera, 
			octant, mOnlyShadowCasters, mLeavePassesInQueue);

		mVisibleNodes.push_back(node);
	}
#endif
	
}
//-----------------------------------------------------------------------
bool OctreeHierarchyInterface::IsNodeVisible(GtpVisibility::HierarchyNode *node) const
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	return static_cast<Octree *>(node)->isOctreeVisible();
#else
	return true;
#endif
}
//-----------------------------------------------------------------------
unsigned int OctreeHierarchyInterface::LastVisited(GtpVisibility::HierarchyNode *node) const
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	return static_cast<Octree *>(node)->lastVisited();
#else
	return 0;
#endif
}
//-----------------------------------------------------------------------
AxisAlignedBox *OctreeHierarchyInterface::GetBoundingBox(GtpVisibility::HierarchyNode *node)
{
	if (node != mSavedNode)
	{
		mSavedNode = node;
	    //static_cast<Octree *>(node)->_getCullBounds(&mBox);
		mBox = static_cast<Octree *>(node)->_getWorldAABB();
	}

	return &mBox;
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::VisualizeCulledNode(GtpVisibility::HierarchyNode *node, 
												   GtpVisibility::CullingType type) const
{
	WireBoundingBox *box = static_cast<Octree *>(node)->getWireBoundingBox();

	if (type == GtpVisibility::FRUSTUM_CULLED)
	{
		box->setMaterial("FrustumCulledNodesMaterial");
	}
	else // type == GtpVisibility::QUERY_CULLED
	{
		box->setMaterial("QueryCulledNodesMaterial");
	}

	dynamic_cast<OctreeSceneManager *>(mSceneManager)->getBoxes()->push_back(box);
}
//-----------------------------------------------------------------------
void OctreeHierarchyInterface::GetNodeGeometryList(GtpVisibility::HierarchyNode *node,	
										   GtpVisibility::GeometryVector *geometryList, 
										   bool includeChildren)
{
	NodeList::const_iterator nodeIt, nodeIt_end;
	nodeIt_end = static_cast<Octree *>(node)->mNodes.end();

	for (nodeIt = static_cast<Octree *>(node)->mNodes.begin(); nodeIt != nodeIt_end; ++nodeIt)
	{
		SceneNodeHierarchyInterface::GetNodeGeometryList(*nodeIt, geometryList, includeChildren);
	}
}

} // namespace Ogre
